
/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.

  FILE:		UdmInt.cpp
  SYNOPSIS:	Udm Interface to different interpretter engines
  CREATED:	08/10/02
  AUTHOR:	Endre Magyari(endre.magyari@vanderbilt.edu)

  CHANGELOG:

  03/28/05 - kalmar
        if (impl) delete[] impl, impl = 0; bug fixed

  02/11/05 - Endre - Fixed getAssociation() to also return the legs of assoc. class objects
  01/28/05 - Himanshu - Adding a equals operator for UdmPseudoObject
  01/27/05 - Himanshu - Defining separate header files for CINT (UdmCint.h) and SWIG (UdmCintSwig.h)
				and adding functions for array attribute getters and setters
  01/11/05 - Himanshu - Added count() functions for functions returning a container of elements
  27/12/02	-	extended the UdmPseudo API with UdmPseudoDataNetwork class, and other initialization classes
  08/10/02	- Added Support for EiC C interpreter. See details 
				at EiC_Interpreter_iface_section



*/



#include <UdmBase.h>
#include <UdmInt.h>
#include <cint_string.h>
#include <UdmCintSwig.h>
#include <Uml.h>
#include <UmlExt.h>
#include <UdmDom.h>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>

using namespace Udm;
using namespace Uml;

//includes from the EiC package,
//which is again ANSI C 



//includes from the Cint package
#include <G__ci.h>						//main header file
extern "C" void G__cpp_setup(void);		//initialize dictionary

//error handling static variables
//for both EiC and Cint interpreters

static string last_error;			//static string to store the last error
static bool was_error = false;		//static bool to indicate an error after the last
									//GetLastError()

//ANSI C struct defined in Eic_interpretted code
typedef struct _eic_upi_UdmPseudoObject
{
	unsigned long dn_id;
	unsigned long ob_id;
} eic_upi_Udm_Pseudo_Object;


//structure to identify an object
struct _upi_o_id {
	unsigned long dnid;
	unsigned long oid;
	_upi_o_id(const unsigned long d, const unsigned long o)
	{
		dnid = d;
		oid = o;
	}
};

//get a UDM Object based on it's Data Network ID and ObjectID
Udm::DataNetwork * GetUdmDataNetwork(unsigned long dn_id)
{
	Udm::DataNetwork * dn = Udm::DataNetwork::GetDnById(dn_id);
	if (dn) return dn;
	throw udm_exception("DataNetwork not found!");
};
Udm::Object GetUdmObject(_upi_o_id id)
{

	Udm::DataNetwork * dn = GetUdmDataNetwork(id.dnid);
	//no check needed - if problems exception is thrown and caught by macro
	Udm::Object o =  dn->ObjectById(id.oid);
	if (o) return o;
	throw udm_exception("Object in DataNetwork not found!");
};


//set an error condition
//because of an exception thrown by UDM::Object API calls
void set_last_error(const char * what)
{
	was_error = true;
	last_error = what;
};



//try / catch macros for getting the object from the ids. exception is thrown & catched if not found
//this also sets the last error to the appropiate text
#define EIC_UPI_TRY(x, y) try {Udm::Object upi_o = GetUdmObject(_upi_o_id(x, y));
#define EIC_UPI_CATCH } catch (const udm_exception& e) {v.cval = EIC_FALSE; set_last_error(e.what());} return v;

//retrieves & checks the id arguments from the EiC stack
#define EIC_UPI_CHECK_ARGS \
	val_t v;\
	unsigned long dnid = arg(0, getargs(), unsigned long);\
	unsigned long oid = arg(1, getargs(), unsigned long);\
	char * att_name = arg(2, getargs(), char *);\
	EIC_UPI_TRY(dnid, oid)\
	

//checks the attribute name
#define EIC_UPI_CHECK_ATTNAME if ((att_name == NULL) || (strlen(att_name) == 0)) {v.cval = EIC_FALSE; return v;};

//retrieves & check the attributes from the EiC stack
#define EIC_UPI_CHECK_ATTARGS \
	val_t v;\
	unsigned long dnid = arg(0, getargs(), unsigned long);\
	unsigned long oid = arg(1, getargs(), unsigned long);\
	char * att_name = arg(2, getargs(), char *);\
	EIC_UPI_TRY(dnid, oid)\
	EIC_UPI_CHECK_ATTNAME\

//checks whether a pointer is an EiC safe Pointer
#define EIC_UPI_ISSAFEPTR(safe_p, eic_val_if_not_safe)\
	if (safe_p.sp == NULL) \
	{\
		/*target is not a safe pointer*/\
		/*refuse to write anything there*/\
		v.cval = eic_val_if_not_safe;\
		return v;\
	}\
	
//copy *safely* a string to a safe pointer allocated in EiC
//refuse to copy, if the pointer is not safe
//copy only to the end of the safe pointer
#define EIC_UPI_STRING_TO_SP(str, safe_p, eic_val_if_not_safe)\
	EIC_UPI_ISSAFEPTR(safe_p, eic_val_if_not_safe)\
	char * ch_p = (char *)safe_p.p;\
	basic_string<char>::const_iterator str_ci = str.begin();\
	for (; str_ci != str.end(); str_ci++)\
	{\
		if (ch_p == (char *) safe_p.ep)\
		{\
			*ch_p = '\0';\
			break;\
		}\
		*ch_p++ = *str_ci;\
	}\
	*ch_p = '\0';\

//gets last error, which was a result of an exception caught 
//in this interface. This function clears the error flag.

#ifdef UDM_EIC_SUPPORT

val_t eic_upi_GetLastError(void)
{
	val_t v;
	if (!was_error) 
	{
		v.cval = EIC_FALSE;
		return v;
	}
	ptr_t safe_p = arg(0, getargs(), ptr_t);
	//copy string to safe pointer
	EIC_UPI_STRING_TO_SP(last_error, safe_p, EIC_TRUE)
	
	was_error = false;
	v.cval = EIC_TRUE;
	return v;

}



//attribute setter functions
val_t eic_upi_SetIntegerAttribute(void)
{
	EIC_UPI_CHECK_ATTARGS;

	long val = arg(3, getargs(), long);

	if (upi_o.SetIntValue(string(att_name), val))
		v.cval = EIC_TRUE;
	else v.cval = EIC_FALSE;

	EIC_UPI_CATCH;
};


val_t eic_upi_SetRealAttribute(void)
{
	EIC_UPI_CHECK_ATTARGS;

	double val = arg(3, getargs(), double);

	if (upi_o.SetRealValue(string(att_name), val))
		v.cval = EIC_TRUE;
	else v.cval = EIC_FALSE;

	EIC_UPI_CATCH;
};

val_t eic_upi_SetBooleanAttribute(void)
{
	EIC_UPI_CHECK_ATTARGS;

	char val = arg(3, getargs(), char);

	if (upi_o.SetBoolValue(string(att_name), val ? true : false))
		v.cval = EIC_TRUE;
	else v.cval = EIC_FALSE;

	EIC_UPI_CATCH;
};

val_t eic_upi_SetStringAttribute(void)
{
	EIC_UPI_CHECK_ATTARGS;

	char * val = arg(3, getargs(), char *);

	if (upi_o.SetStrValue(string(att_name), val))
		v.cval = EIC_TRUE;
	else v.cval = EIC_FALSE;

	EIC_UPI_CATCH;
};

//atribute getter functions
val_t eic_upi_GetStringAttribute(void)
{
	EIC_UPI_CHECK_ATTARGS;

	string att_val;
	if (!upi_o.GetStrValue(att_name, att_val))
	{
		v.cval = EIC_FALSE;
		return v;
	}

	ptr_t safe_p = arg(3, getargs(), ptr_t);
	//copy string to safe pointer
	EIC_UPI_STRING_TO_SP(att_val, safe_p, EIC_TRUE)
	
	EIC_UPI_CATCH;
};

//attribute getter functions
val_t eic_upi_GetRealAttribute(void)
{
	EIC_UPI_CHECK_ATTARGS;

	double att_val;
	if (!upi_o.GetRealValue(att_name, att_val))
	{
		v.cval = EIC_FALSE;
		return v;
	}

	double * dbl_p = arg(3, getargs(), double *);
	*dbl_p = att_val;

	EIC_UPI_CATCH;
};

//attribute getter functions
val_t eic_upi_GetIntegerAttribute(void)
{
	EIC_UPI_CHECK_ATTARGS;

	long att_val;
	if (!upi_o.GetIntValue(att_name, att_val))
	{
		v.cval = EIC_FALSE;
		return v;
	}

	long * long_p = arg(3, getargs(), long *);
	*long_p = att_val;

	EIC_UPI_CATCH;
};


val_t eic_upi_GetBooleanAttribute(void)
{
	EIC_UPI_CHECK_ATTARGS;

	bool att_val;
	if (!upi_o.GetBoolValue(att_name, att_val))
	{
		v.cval = EIC_FALSE;
		return v;
	}

	char * char_p = arg(3, getargs(), char *);
	*char_p = att_val ? EIC_TRUE : EIC_FALSE;

	EIC_UPI_CATCH;
};

//initialization stuff for EiC interpreter

void eic_init(char * eic_switches, void (*msg_handler)(char *))
{
	/* initiate EiC */
	EiC_init_EiC();

	/* add in the standard C library */
	stdClib();

	/* pass EiC the following command line switches */
	EiC_switches(eic_switches);

	/* you can also set the function used to display EiC's error and warning
		 messages */
	EiC_setMessageDisplay(msg_handler);

	//adding as built-ins
	EiC_add_builtinfunc("GetLastError",eic_upi_GetLastError);
	EiC_add_builtinfunc("SetIntegerAttribute",eic_upi_SetIntegerAttribute);
	EiC_add_builtinfunc("SetRealAttribute",eic_upi_SetRealAttribute);
	EiC_add_builtinfunc("SetBooleanAttribute",eic_upi_SetBooleanAttribute);
	EiC_add_builtinfunc("SetStringAttribute",eic_upi_SetStringAttribute);
	EiC_add_builtinfunc("GetIntegerAttribute",eic_upi_GetIntegerAttribute);
	EiC_add_builtinfunc("GetRealAttribute",eic_upi_GetRealAttribute);
	EiC_add_builtinfunc("GetBooleanAttribute",eic_upi_GetBooleanAttribute);
	EiC_add_builtinfunc("GetStringAttribute",eic_upi_GetStringAttribute);


	//for convenience, include all the common libraries
	EiC_parseString("#include <stdio.h>");
	EiC_parseString("#include <stdlib.h>");
	EiC_parseString("#include <string.h>");

	//bool type suport, and constants
	EiC_parseString("typedef char bool;");
	EiC_parseString("#define true 0x01");
	EiC_parseString("#define false 0x00");
	
			

	//UdmPseudoObject structure for identifying a Udm::Object
	EiC_parseString("\
		typedef struct _UdmPseudoObject{\
			unsigned long dn_id;\
			unsigned long ob_id;\
		} UdmPseudoObject;");


	//adding prototypes for built-ins
	
	/*	bool GetLastError(char[] buffer);
	 *		- gets the last error, if any. If there was an error, the return 
	 *			value is true, otherwise is false; If there was an error, a description
	 *			is copied in the buffer. Only safe pointers are accepted. Only allocated
	 *			number of characters - 1 are copied.
	 */
	EiC_parseString("bool GetLastError(char[]);");


	/*	Attribute value setters - accessing the objects with pair of ids
	 *	If the return value is true, the operation completed OK.
	 *	If the return value is false, and GetLastError() is false, then 
	 *	there was no such attribute.
	 *  If the return value is false, and GetLastError() is true, then 
	 *  either the datanetwork or the object was not found, or for some 
	 *  other reason a udm_exception was thrown. The error message is a description
	 *  of the exception.
	 */
	EiC_parseString("bool SetIntegerAttribute(const unsigned long, const unsigned long, const char *, long);");
	EiC_parseString("bool SetRealAttribute(const unsigned long, const unsigned long, const char *, double);");
	EiC_parseString("bool SetBooleanAttribute(const unsigned long, const unsigned long, const char *, bool);");
	EiC_parseString("bool SetStringAttribute(const unsigned long, const unsigned long, const char *, const char *);");

	/* Attribute value getters - accessing the objects with pair of ids
	 * Same rules apply, in addition, in particular for GetStringAttribute 
	 * the size of the buffer is always checked. Only safe pointers are accepted. Only
	 * allocated number of characters - 1 are copied.
	 */
	EiC_parseString("bool GetIntegerAttribute(const unsigned long, const unsigned long, const char *, long *);");
	EiC_parseString("bool GetRealAttribute(const unsigned long, const unsigned long, const char *, double *);");
	EiC_parseString("bool GetBooleanAttribute(const unsigned long, const unsigned long, const char *, bool *);");
	EiC_parseString("bool GetStringAttribute(const unsigned long, const unsigned long, const char *, char *);");

	/*
	 * Attribute accessors based on UdmPseudoObject structs.
	 * Same rules apply.
	 */

	EiC_parseString("bool SetIntVal(UdmPseudoObject *o, const char *n, long v){return SetIntegerAttribute(o->dn_id, o->ob_id, n, v);}");
	EiC_parseString("bool SetRealVal(UdmPseudoObject *o, const char *n, double v){return SetRealAttribute(o->dn_id, o->ob_id, n, v);}");
	EiC_parseString("bool SetBoolVal(UdmPseudoObject *o, const char *n, bool v){return SetBooleanAttribute(o->dn_id, o->ob_id, n, v);}");
	EiC_parseString("bool SetStrVal(UdmPseudoObject *o, const char *n, const char * v){return SetStringAttribute(o->dn_id, o->ob_id, n, v);}");
	EiC_parseString("bool GetIntVal(UdmPseudoObject *o, const char *n, long * v){return GetIntegerAttribute(o->dn_id, o->ob_id, n, v);}");
	EiC_parseString("bool GetRealVal(UdmPseudoObject *o, const char *n, double * v){return GetRealAttribute(o->dn_id, o->ob_id, n, v);}");
	EiC_parseString("bool GetBoolVal(UdmPseudoObject *o, const char *n, bool * v){return GetBooleanAttribute(o->dn_id, o->ob_id, n, v);}");
	EiC_parseString("bool GetStrVal(UdmPseudoObject *o, const char *n, char * v){return GetStringAttribute(o->dn_id, o->ob_id, n, v);}");


};

static set<string> eic_global_vars;

void eic_add_object(Udm::Object o, string o_name)
{

	//setting up a new object for object o;
	basic_string<char>::iterator i;
	for(i = o_name.begin(); i != o_name.end(); i++) 
	{
		if(!isalnum(*i)) *i = '_';
	}

	if (o_name.empty())
		throw udm_exception("Can not set up an EiC PseudoObject with an empty name!");

	//insert & check for potential name conflict
	pair <set<string>::iterator, bool> ins_res = eic_global_vars.insert(o_name);

	if (!ins_res.second)
		throw udm_exception("Global PseudoObject variable " + o_name + " is already defined. First clear it!");



	//obtain the pair of Ids
	unsigned long dn_id = o.__impl()->__getdn()->uniqueId();
	unsigned long o_id = o.__impl()->uniqueId();

	//create the object
	EiC_parseString("UdmPseudoObject _%s;", o_name.c_str());
	EiC_parseString("UdmPseudoObject * %s = &_%s;", o_name.c_str(), o_name.c_str());

	//set the dn and ob id's
	EiC_parseString("%s->dn_id = %d; %s->ob_id = %d;", o_name.c_str(), dn_id,  o_name.c_str(), o_id);
	
	
};


void eic_rem_object(string name, bool throw_not_found)
{
	basic_string<char>::iterator i;
	for(i = name.begin(); i != name.end(); i++) 
	{
		if(!isalnum(*i)) *i = '_';
	}

	if (name.empty())
		throw udm_exception("Can not remove an EiC PseudoObject with an empty name!");

	if (eic_global_vars.find(name) != eic_global_vars.end())
	{
		eic_global_vars.erase(name);	
		EiC_parseString((char *)(string(string(":rm ") + name).c_str()));
		EiC_parseString((char *)(string(string(":rm _") + name).c_str()));


	}
	else if (throw_not_found)
		throw udm_exception("Global PseudoObject variable " + name + " is not defined.");
		
}

void eic_clear()
{
	set<string>::iterator i = eic_global_vars.begin();
	for (i = eic_global_vars.begin(); i != eic_global_vars.end(); i++)
	{
		EiC_parseString((char *)(string(string(":rm ") + *i).c_str()));
		EiC_parseString((char *)(string(string(":rm _") + *i).c_str()));

	}

	eic_global_vars.clear();
}

void eic_exec(string code)
{

	EiC_parseString((char *)code.c_str());
};


//list all  objects add as global variables
void eic_prompt_sh_objects()
{
	cout << "Defined global UdmPseudoObject variables: " << endl;
	set<string>::iterator i = eic_global_vars.begin();
	for (i = eic_global_vars.begin(); i != eic_global_vars.end(); i++)
	{
		fprintf(stdout,  "UdmPseudoObject %s;\n" , i->c_str());
	}
};

char *inbuf;




void eic_ipretter()
{
	
	
	fprintf(stdout, "============================================================\n");
	fprintf(stdout, "Udm ANSI C interpretter running EiC" );
	fprintf(stdout, "============================================================\n");

	eic_prompt_sh_objects();

	unsigned char * line;
	code_t code;
	int h = EiC_interActive;

	while(true)
	{

		line = EiC_readline("Udm EiC >");
	
		if (line)
		{
			EiC_add_history((char *)line);

			//surpress :exit command, and add for convenience :q vi style exit
			if  ((!strcmp((const char *)line, ":q")) || (!strcmp((char *)line, ":exit")) || (!strcmp((char *)line, ":quit"))) break;

			//show all the global object variables
			if	(!strcmp((const char *)line, ":?")) 
			{
				eic_prompt_sh_objects();
				continue;
			}

			EiC_interActive = 0;  /* turn off interactive mode */
			EiC_prolineString((char *)line);


			code = EiC_ENV->CODE;
			EiC_ENV->CODE.nextinst = EiC_ENV->CODE.binst = 0;
			EiC_ENV->CODE.inst = NULL;
			EiC_ENV->CODE.prev = &code;
			/*  sp = EiC_ENV->sp;*/

			runEiC();

			/* rem: runEiC will call free_env_code */
			EiC_ENV->CODE = code;

			EiC_interActive = h;
			free(line);
		}
	}
};

#endif //UDM_EIC_SUPPORT
//Cint API functions


//init the CInt C/C++ interpreter
void cint_init(char * cint_switches, void (*msg_handler)(char *))
{
	//add UDM Pseudo API to the dictionary
	
#ifdef _WIN32
	G__set_p2fsetup(G__cpp_setup);
#else
	//g++ complains about this line, so
	//cast is needed, ANSI C does not support 
	//the implicit cast of
	//void (*)(...) to void (*)()
	G__set_p2fsetup((void (*)())G__cpp_setup);
#endif

	//compute init string
	string init_str = "cint";
	if (cint_switches && strlen(cint_switches))
	{
		init_str += " ";
		init_str += cint_switches;
	}

	//init the library
	G__init_cint(init_str.c_str());
};

//destroy the interpretter
void cint_destroy()
{
	//destroy
	G__scratch_all();
};


//add an object o to the existing code block
//two objects with the same name can not be declared in the {same block}.
void cint_add_object(Udm::Object o, string name)
{
	char buff[200];
	
	sprintf(buff,"UdmPseudoObject %s(%ld, %ld);", name.c_str(),  o.__impl()->__getdn()->uniqueId(), o.__impl()->uniqueId());
	G__exec_text(buff);

};


//starts interactive mode
void cint_ipretter()
{
	
	char* line;
	unsigned char opened_pars = 0;
	int i;
	
	//create an initial prompt
	string prompt = "Udm CInt >";
	
	//what will be executed
	string code;

	while ((line = G__input((char *)prompt.c_str())))
	{

		//change the ending new-line char to string-ending \0
		*(line + strlen(line) - 1) = '\0';

		//surpress :exit command, and add for convenience :q vi style exit
		if  ((!strcmp((const char *)line, ":q")) || (!strcmp((char *)line, ":exit")) || (!strcmp((char *)line, ":quit"))) break;

		//add the line to code
		code += line;
		

		char * line_c = line;
		bool inside_par = false;
		bool escaped = false;

		
		while (*line_c != '\0')
		{
			switch (*line_c++)
			{
			case '"':
				if (!escaped) inside_par = !inside_par;
				escaped = false;
				break;
			
			case '{':
				if (!inside_par) opened_pars++;
				escaped = false;
				break;
			
			case '}':
				if (!inside_par) opened_pars--;
				escaped = false;
				break;
		
			case '\\':
				escaped = true;
				break;
			
			default:
				escaped = false;
				break;
			};
		}
	
		
		if (!opened_pars)
		{
			
			//executing code
			G__value res = G__exec_text(code.c_str());

			//getting the results
			int res_int = G__int(res);
			double res_dbl = G__double(res);

			//print the return value of the expression
			if ((double)res_int != res_dbl)
				printf(">\t: %lf\n", res_dbl);
			else 
				printf(">\t: %d\n", res_int);

			//clear the code
			code.erase();
		}

		//make-up the prompt 
		prompt = "Udm CInt >";
		i = 0;
		while ( i++ < opened_pars) prompt += "{>";

	}
};



int cint_int_calc(string code)
{
	return G__int(G__calc(code.c_str()));
};

double cint_double_calc(string code)
{
	return G__double(G__calc(code.c_str()));
};

int cint_int_eval(string code)
{
	return G__int(G__exec_text(code.c_str()));
};

double cint_double_eval(string code)
{
	return G__double(G__exec_text(code.c_str()));
};


char* cint_load_text(const char *namedmacro)
{
	return G__load_text(namedmacro);
};

int cint_loadfile(const char *filename)
{
	return G__loadfile(filename);
};

int cint_unloadfile(const char *filename)
{
	return G__unloadfile(filename);
}



//Cint stuff - implementation of the CInt UdmPseudoObject interface
//these objects will be known to the CIint interpretter context,
//as prceompiled code
UdmPseudoObjectS::UdmPseudoObjectS(int length)
{
	if (length)
	{
		impl = new UdmPseudoObject[length];
	}
	else impl = NULL;

	this->length = length;
	
};

UdmPseudoObjectS& UdmPseudoObjectS::operator=(const UdmPseudoObjectS &frm)
{
	delete[] impl;
	length = frm.length;
	if (length) impl = new UdmPseudoObject[length];
	
	for (int i = 0; i < length; i++)
	{
		UdmPseudoObject * to = impl + i;
		*to = *(frm.impl + i);
	}
	return *this;
};

UdmPseudoObjectS::UdmPseudoObjectS(const UdmPseudoObjectS &frm)
{
	length = frm.length;
	if (length) impl = new UdmPseudoObject[length];
	
	for (int i = 0; i < length; i++)
	{
		UdmPseudoObject * to = impl + i;
		*to = *(frm.impl + i);
	}
};

UdmPseudoObject& UdmPseudoObjectS::operator[](const int index) const
{
	int index_ok = index;
	((UdmPseudoObjectS*)this)->overflow = false;		//hack because of const-ness
	if (index_ok > length - 1) 
	{
		index_ok = length - 1;
		((UdmPseudoObjectS*)this)->overflow = true;		//hack because of const-ness
	}

	return *(impl + index_ok);
};
	
void UdmPseudoObjectS::SetAt(const int index, const UdmPseudoObject& item)
{
	int index_ok = index;
	overflow = false;
	if (index_ok > length - 1)
	{
		index_ok = length - 1;
		overflow = true;
	}
	*(impl + index_ok) = item;

};

int UdmPseudoObjectS::GetLength() const
{
	return length;
};


UdmPseudoObjectS::~UdmPseudoObjectS()
{
	if (impl) delete[] impl;
	impl = NULL;
};

bool UdmPseudoObjectS::operator !() const { return length == 0; }
UdmPseudoObjectS::operator bool()  const { return  length != 0; }


AssociationInfo::AssociationInfo(const char *src, const char *dst, const char *assoc)
{
	this->SrcRolename = src;
	this->DstRolename = dst;
	this->assoc_class = assoc;
};

AssociationInfo::AssociationInfo(cint_string src, cint_string dst, cint_string assoc)
{
	SrcRolename = src;
	DstRolename = dst;
	assoc_class = assoc;
};

CompositionInfo::CompositionInfo(const char *pr, const char *cr)
{
	parentRole = pr;
	childRole = cr;
};

CompositionInfo::CompositionInfo(cint_string pr, cint_string cr)
{
	parentRole = pr;
	childRole = cr;
};



/*
class UdmPseudoObject
{
	unsigned long ob_id;
	unsigned long dn_id;
}*/	

#define CINT_TRY(x, y) try {Udm::Object upi_o = GetUdmObject(_upi_o_id(x, y))
#define CINT_TRY2(x1, y1, x2, y2) try {Udm::Object upi_o1 = GetUdmObject(_upi_o_id(x1, y1)); Udm::Object upi_o2 = GetUdmObject(_upi_o_id(x2, y2));
#define CINT_CATCH } catch (const udm_exception& e) {set_last_error(e.what()); return false;} catch(...) {set_last_error("Unexpected exception caught."); return false;}
#define CINT_DN_TRY(x) try {Udm::DataNetwork * dn = GetUdmDataNetwork(x)
#define CINT_DN_CATCH } catch (const udm_exception& e) {set_last_error(e.what()); return false;} catch(...) {set_last_error("Unexpected exception caught."); return false;}


UdmPseudoObject::UdmPseudoObject()
{
	this->dn_id = 0;
	this->ob_id = 0;
};


UdmPseudoObject::UdmPseudoObject(unsigned long dn_id, unsigned long ob_id)
{
	this->dn_id = dn_id;
	this->ob_id = ob_id;
/*
	try
	{
		Udm::Object upi_o = GetUdmObject(_upi_o_id(dn_id, ob_id));
		upi_o.__impl()->clone();
	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
	};
	*/
}
UdmPseudoObject::~UdmPseudoObject()
{
}

UdmPseudoObject& UdmPseudoObject::operator=(const UdmPseudoObject& frm)
{
	this->dn_id = frm.dn_id;
	this->ob_id = frm.ob_id;
/*
	try
	{
		Udm::Object upi_o = GetUdmObject(_upi_o_id(dn_id, ob_id));
		upi_o.__impl()->clone();
	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
	};

*/
	return *this;
};

unsigned long UdmPseudoObject::_ob_id() const
{
	return ob_id;
};
unsigned long UdmPseudoObject::_dn_id() const
{
	return dn_id;
};


// Himanshu: Adding a equals operator for UdmPseudoObject
UDM_DLL bool UdmPseudoObject::operator ==(const UdmPseudoObject& to)
{
	CINT_TRY(dn_id, ob_id);
	CINT_TRY(to.dn_id, to.ob_id);
	return ((this->ob_id == to.ob_id) && (this->dn_id == to.dn_id));
	CINT_CATCH;
	CINT_CATCH;
}

//error handling
bool UdmPseudoObject::GetLastError(cint_string& buffer)
{
	if (was_error)
	{
		//buffer.CopyFrom(last_error.c_str());
		buffer = last_error.c_str();
		was_error = false;
		return true;
	} else return false;
};

//type
bool UdmPseudoObject::type(cint_string &value) const
{
	CINT_TRY(dn_id, ob_id);

	string str = upi_o.type().name();
	value = str.c_str();	//this will reallocate !!
	return true;

	CINT_CATCH;
};

bool UdmPseudoObject::stereotype(cint_string &value) const
{
	CINT_TRY(dn_id, ob_id);

	string str = upi_o.type().stereotype();
	value = str.c_str();	//this will reallocate !!
	
	return true;

	CINT_CATCH;
};

bool UdmPseudoObject::uniqueId(UdmLong &value) const
{
	CINT_TRY(dn_id, ob_id);
	value.longVal = upi_o.uniqueId();
	return true;
	CINT_CATCH;
}

//IMPL API
bool UdmPseudoObject::getParent(const char * prole, UdmPseudoObject &value) const
{
	CINT_TRY(dn_id, ob_id);
	//obtain parentrole from prole
	CompositionParentRole cpr;
	if (prole)
	{
		set<CompositionParentRole> proles = AncestorCompositionPeerParentRoles(upi_o.type());
		set<CompositionParentRole>::iterator proles_i = proles.begin();
		while (proles_i != proles.end())
		{
			if (!strcmp(prole, ((string)proles_i->name()).c_str()))
					break;
			proles_i++;
		} 

		if (proles_i == proles.end()) throw udm_exception(string("No such parentrole: ") + prole);
		cpr = *proles_i;
	}
	//getparent does the cloning
	Udm::Object par = upi_o.__impl()->getParent(cpr);
	if (par)
	{
		UdmPseudoObject upo(par.__impl()->__getdn()->uniqueId(), par.__impl()->uniqueId());
		value = upo;
	}else value = UdmPseudoObject();

	return true;

	CINT_CATCH;
};

bool UdmPseudoObject::setParent(UdmPseudoObject &parent, const char * prole)
{
	CINT_TRY(dn_id, ob_id);
	
	//the new parent
	Udm::Object new_parent = GetUdmObject(_upi_o_id(parent.dn_id, parent.ob_id));

	CompositionParentRole cpr;
	//obtain parentrole from prole
	if(prole)
	{
		set<CompositionParentRole> proles = AncestorCompositionPeerParentRoles(upi_o.type());
		set<CompositionParentRole>::iterator proles_i = proles.begin();
		while (proles_i != proles.end())
		{
			if (!strcmp(prole, ((string)proles_i->name()).c_str()))
				break;
			proles_i++;
		}

		if (proles_i == proles.end()) throw udm_exception(string("No such parentrole: ") + prole);
		cpr = *proles_i;
	}


	upi_o.__impl()->setParent(new_parent.__impl(), cpr);
	return true;
	
	CINT_CATCH;
};

bool UdmPseudoObject::detach()
{
	CINT_TRY(dn_id, ob_id);
	upi_o.__impl()->detach();
	return true;
	CINT_CATCH;
};
	
bool UdmPseudoObject::getChildrenCount(const char * crole, const char *k, UdmLong &value) const // HN1
{
	CINT_TRY(dn_id, ob_id);
	CompositionChildRole ccr;//it will be NullObject by default
	Class kind;

	if (crole)
	{
		set<CompositionChildRole> croles = AncestorCompositionPeerChildRoles(upi_o.type());
		set<CompositionChildRole>::iterator croles_i = croles.begin();
		while (croles_i != croles.end())
		{
			if (!strcmp(crole, ((string)croles_i->name()).c_str()))
				break;
			croles_i++;
		}

		if (croles_i == croles.end()) throw udm_exception(string("No such childrole: ") + crole);;
		ccr = *croles_i;
	}

	if (k)
	{
		kind =  classByName(upi_o.__impl()->__getdn()->GetRootMeta(), k) ;
		if (!kind) throw udm_exception(string("No such class: ") + k);;
	}

	vector<ObjectImpl*> children = upi_o.__impl()->getChildren(ccr, kind);

	//impls needs to be freed!
	vector<ObjectImpl*>::iterator children_i = children.begin();
	while (children_i != children.end())
		Udm::Object child(*children_i++);	//so, it will be freed
	
	
	value.longVal = children.size();

	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::getChildren(const char * crole, const char *k, UdmPseudoObjectS &value) const
{
	CINT_TRY(dn_id, ob_id);
	CompositionChildRole ccr;//it will be NullObject by default
	Class kind;

	if (crole)
	{
		set<CompositionChildRole> croles = AncestorCompositionPeerChildRoles(upi_o.type());
		set<CompositionChildRole>::iterator croles_i = croles.begin();
		while (croles_i != croles.end())
		{
			if (!strcmp(crole, ((string)croles_i->name()).c_str()))
				break;
			croles_i++;
		}

		if (croles_i == croles.end()) throw udm_exception(string("No such childrole: ") + crole);;
		ccr = *croles_i;
	}

	if (k)
	{
		kind =  classByName(upi_o.__impl()->__getdn()->GetRootMeta(), k) ;
		if (!kind) throw udm_exception(string("No such class: ") + k);;
	}

	vector<ObjectImpl*> children = upi_o.__impl()->getChildren(ccr, kind);
	vector<ObjectImpl*>::iterator children_i = children.begin();
	int i = 0;
	while (children_i != children.end())
	{
		Udm::Object child(*children_i++);	//so, it will be freed
		UdmPseudoObject upo(child.__impl()->__getdn()->uniqueId(), child.uniqueId());
		value.SetAt(i++ ,upo);
	}
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::setChildren(const char * crole, const UdmPseudoObjectS &children)
{
	CINT_TRY(dn_id, ob_id);

	vector<ObjectImpl*> children_vec;

	for (int i = 0; i < children.GetLength(); i++)
	{
		UdmPseudoObject child = children[i];
		Udm::Object new_child = GetUdmObject(_upi_o_id(child.dn_id, child.ob_id));
		children_vec.push_back(new_child.__impl());
	};

	CompositionChildRole ccr;

	if (crole)
	{
		set<CompositionChildRole> croles = AncestorCompositionPeerChildRoles(upi_o.type());
		set<CompositionChildRole>::iterator croles_i = croles.begin();
		while (croles_i != croles.end())
		{
			if (!strcmp(crole, ((string)croles_i->name()).c_str()))
				break;
			croles_i++;
		}

		if (croles_i == croles.end()) throw udm_exception(string("No such childrole: ") + crole);;
		ccr = *croles_i;
	}

	upi_o.__impl()->setChildren(ccr, children_vec);
	return true;
	CINT_CATCH;
};




//associations
bool UdmPseudoObject::getAssociationCount(const char* ass_role, UdmLong& value, int mode) const // HN2
{
	CINT_TRY(dn_id, ob_id);
	
	if (!ass_role) return false;
	set<AssociationRole> aroles = AncestorAssociationTargetRoles(upi_o.type());
	
	Uml::Association assoc = GetAncestorAssociation(upi_o.type());
	if (assoc)
	{
		set<AssociationRole> legs = assoc.roles();
		aroles.insert(legs.begin(), legs.end());
	}

	
	set<AssociationRole>::iterator aroles_i = aroles.begin();
	while (aroles_i != aroles.end())
	{
		if (!strcmp(ass_role, ((string)aroles_i->name()).c_str()))
			break;
		aroles_i++;
	}

	
	if (aroles_i == aroles.end()) throw udm_exception(string("No such associationrole: ") + ass_role);;

	vector<ObjectImpl*> ass_impls = upi_o.__impl()->getAssociation(*aroles_i, mode);
	value.longVal = ass_impls.size();

	vector<ObjectImpl*>::iterator ass_impls_i = ass_impls.begin();
	while (ass_impls_i != ass_impls.end())
		Udm::Object ass_object(*ass_impls_i++);	//so, it will be freed


	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::getAssociation(const char* ass_role, UdmPseudoObjectS &value, int mode) const
{
	CINT_TRY(dn_id, ob_id);
	
	if (!ass_role) return false;
	set<AssociationRole> aroles = AncestorAssociationTargetRoles(upi_o.type());

	Uml::Association assoc = GetAncestorAssociation(upi_o.type());
	if (assoc)
	{
		set<AssociationRole> legs = assoc.roles();
		aroles.insert(legs.begin(), legs.end());
	}

	set<AssociationRole>::iterator aroles_i = aroles.begin();
	while (aroles_i != aroles.end())
	{
		if (!strcmp(ass_role, ((string)aroles_i->name()).c_str()))
			break;
		aroles_i++;
	}
	if (aroles_i == aroles.end()) throw udm_exception(string("No such associationrole: ") + ass_role);;

	vector<ObjectImpl*> ass_impls = upi_o.__impl()->getAssociation(*aroles_i, mode);
	vector<ObjectImpl*>::iterator ass_impls_i = ass_impls.begin();
	int i = 0;
	while (ass_impls_i != ass_impls.end())
	{
		Udm::Object ass_object(*ass_impls_i++);	//so, it will be freed
		UdmPseudoObject upo(ass_object.__impl()->__getdn()->uniqueId(), ass_object.uniqueId());
		value.SetAt(i++, upo);
	}
	return true;

	CINT_CATCH;
};
bool UdmPseudoObject::setAssociation(const char * ass_role, const UdmPseudoObjectS &value, int mode)
{
	CINT_TRY(dn_id, ob_id);
	if (!ass_role) return false;
	Uml::Class cl = upi_o.type();
	Uml::Association cl_assoc = Uml::GetAncestorAssociation(cl);
	set<AssociationRole> aroles;

	if (cl_assoc)
		aroles = cl_assoc.roles();
	else
		aroles = AncestorAssociationTargetRoles(upi_o.type());

	set<AssociationRole>::iterator aroles_i = aroles.begin();
	while (aroles_i != aroles.end())
	{
		if (!strcmp(ass_role, ((string)aroles_i->name()).c_str()))
			break;
		aroles_i++;
	}
	if (aroles_i == aroles.end()) throw udm_exception(string("No such associationrole: ") + ass_role);

	vector<ObjectImpl*> new_ass_impls;
	for (int i = 0; i < value.GetLength(); i++)
	{
		UdmPseudoObject ass_obj = value[i];
		Udm::Object new_ass_obj = GetUdmObject(_upi_o_id(ass_obj.dn_id, ass_obj.ob_id));
		//TBD: this need to be released() somewhere ...
		new_ass_impls.push_back(new_ass_obj.__impl()->clone());
	};

	upi_o.__impl()->setAssociation(*aroles_i,new_ass_impls, mode);

	

	return true;


	CINT_CATCH;
};



//attribute setters
bool UdmPseudoObject::SetIntVal(const char * name, __int64 value)
{
	CINT_TRY(dn_id, ob_id);
	bool ret = upi_o.SetIntValue(name, value);
	if (!ret) throw udm_exception(string("No such attribute: ") + name);
	return true;

	CINT_CATCH;
};
bool UdmPseudoObject::SetRealVal(const char * name, double value)
{
	CINT_TRY(dn_id, ob_id);
	bool ret = upi_o.SetRealValue(name, value);
	if (!ret) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};
bool UdmPseudoObject::SetStrVal(const char * name, const char *  value)
{
	CINT_TRY(dn_id, ob_id);
	bool ret = upi_o.SetStrValue(name, value);
	if (!ret) throw udm_exception(string("No such attribute: ") + name);
	return true;
	
	CINT_CATCH;
};
bool UdmPseudoObject::SetBoolVal(const char * name, bool value)
{
	CINT_TRY(dn_id, ob_id);
	bool ret = upi_o.SetBoolValue(name, value);
	if (!ret) throw udm_exception(string("No such attribute: ") + name);
	return true;
	
	CINT_CATCH;
};

//attribute getters
bool UdmPseudoObject::GetIntVal(const char * name, UdmLong &value)
{
	// Himanshu: Commenting out as now UdmLongs hold __int64 integers
	/*CINT_TRY(dn_id, ob_id);
	__int64 val64;
	bool res = upi_o.GetIntValue(name, val64);
	value.longVal = (long) val64;
	if (!res) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;*/

	CINT_TRY(dn_id, ob_id);
	bool res = upi_o.GetIntValue(name, value.longVal);
	if (!res) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};
bool UdmPseudoObject::GetRealVal(const char * name, UdmDouble &value)
{
	CINT_TRY(dn_id, ob_id);
	bool res =  upi_o.GetRealValue(name, value.doubleVal);
	
	if (!res) throw udm_exception(string("No such attribute: ") + name);
	return true;
	
	CINT_CATCH;
};
bool UdmPseudoObject::GetStrVal(const char * name,  cint_string & value)
{
	CINT_TRY(dn_id, ob_id);
	string stl_value;
	if (upi_o.GetStrValue(name, stl_value))
	{
		//value.CopyFrom(stl_value.c_str());
		value = stl_value.c_str();
		return true;
	}
	throw udm_exception(string("No such attribute: ") + name);
	return false;
	CINT_CATCH;
};
bool UdmPseudoObject::GetBoolVal(const char * name, UdmBool &value)
{
	CINT_TRY(dn_id, ob_id);
	bool val;
	bool retval = upi_o.GetBoolValue(name, val);

	value.boolVal = val;
	if (!retval) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};

// Himanshu: integer array attribute getter
bool UdmPseudoObject::GetIntValues(const char * name, vector<UdmLong>& values)
{
	// Himanshu: Commenting out as now UdmLongs hold __int64 integers
	/*CINT_TRY(dn_id, ob_id);
	std::vector<__int64> val64;
	bool res = upi_o.GetIntValues(name, val64);
	std::vector<UdmLong> val;
	std::vector<__int64>::const_iterator it = val64.begin();
	while(it != val64.end()) {
		UdmLong newVal((long) *it);
		val.push_back(newVal);
		it++;
	}
	values = val;
	if (!res) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;*/

	CINT_TRY(dn_id, ob_id);
	std::vector<__int64> val64;
	bool res = upi_o.GetIntValues(name, val64);
	std::vector<UdmLong> val;
	std::vector<__int64>::const_iterator it = val64.begin();
	while(it != val64.end()) {
		UdmLong newVal(*it);
		val.push_back(newVal);
		it++;
	}
	values = val;
	if (!res) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};

// Himanshu: double array attribute getter
bool UdmPseudoObject::GetRealValues(const char * name, vector<double>& values)
{
	CINT_TRY(dn_id, ob_id);
	bool res =  upi_o.GetRealValues(name, values);
	if (!res) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};

// Himanshu: string array attribute getter
bool UdmPseudoObject::GetStrValues(const char * name, vector<cint_string>& values)
{
	CINT_TRY(dn_id, ob_id);
	vector<string> stl_values;
	if (upi_o.GetStrValues(name, stl_values))
	{
		vector<string>::const_iterator it = stl_values.begin();
		while(it != stl_values.end()) {
			values.push_back((*it).c_str());
			it++;
		}
		return true;
	}
	throw udm_exception(string("No such attribute: ") + name);
	return false;
	CINT_CATCH;
};

// Himanshu: bool array attribute getter
bool UdmPseudoObject::GetBoolValues(const char * name, vector<bool>& values)
{
	CINT_TRY(dn_id, ob_id);
	bool retval = upi_o.GetBoolValues(name, values);
	if (!retval) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};

// Himanshu: integer array attribute setter
bool UdmPseudoObject::SetIntValues(const char * name, vector<UdmLong> values)
{
	// Himanshu: Commenting out as now UdmLongs hold __int64 integers
	/*CINT_TRY(dn_id, ob_id);
	std::vector<__int64> vals64;
	std::vector<UdmLong>::const_iterator it = values.begin();
	while(it != values.end()) {
		vals64.push_back((__int64) ((*it).longVal));
		it++;
	}
	bool ret = upi_o.SetIntValues(name, vals64);
	if (!ret) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;*/


	CINT_TRY(dn_id, ob_id);
	std::vector<__int64> vals64;
	std::vector<UdmLong>::const_iterator it = values.begin();
	while(it != values.end()) {
		vals64.push_back((*it).longVal);
		it++;
	}
	bool ret = upi_o.SetIntValues(name, vals64);
	if (!ret) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};

// Himanshu: double array attribute setter
bool UdmPseudoObject::SetRealValues(const char * name, vector<double> values)
{
	CINT_TRY(dn_id, ob_id);
	bool ret = upi_o.SetRealValues(name, values);
	if (!ret) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};

// Himanshu: string array attribute setter
bool UdmPseudoObject::SetStrValues(const char * name, vector<cint_string> values)
{
	CINT_TRY(dn_id, ob_id);
	vector<string> stl_values;
	vector<cint_string>::const_iterator it = values.begin();
	while(it != values.end()) {
		stl_values.push_back((*it).buffer());
		it++;
	}
	bool ret = upi_o.SetStrValues(name, stl_values);
	if (!ret) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};

// Himanshu: bool array attribute setter
bool UdmPseudoObject::SetBoolValues(const char * name, vector<bool> values)
{
	CINT_TRY(dn_id, ob_id);
	bool ret = upi_o.SetBoolValues(name, values);
	if (!ret) throw udm_exception(string("No such attribute: ") + name);
	return true;
	CINT_CATCH;
};

//TOMI -- to be done later on
bool UdmPseudoObject::GetAdjacentObjectsCount(UdmLong &value) // HN3
{
	CINT_TRY(dn_id, ob_id);
	set<Udm::Object> ret = upi_o.GetAdjacentObjects();
	value.longVal = ret.size();
	
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetAdjacentObjects(UdmPseudoObjectS &value)
{
	CINT_TRY(dn_id, ob_id);
	set<Udm::Object> ret = upi_o.GetAdjacentObjects();
	set<Udm::Object>::iterator ret_i = ret.begin();
	int i = 0;
	while (ret_i != ret.end())
		value.SetAt(i++, UdmPseudoObject(upi_o.__impl()->__getdn()->uniqueId(), upi_o.__impl()->uniqueId()));
	
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetAdjacentObjectsCount(const char * dst_type, UdmLong &value) // HN4
{
	CINT_TRY(dn_id, ob_id);
	
	Uml::Class kind = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), dst_type);
	if (!kind) throw udm_exception(string("No such kind: ") + dst_type);

	set<Udm::Object> ret = upi_o.GetAdjacentObjects(kind);
	value.longVal = ret.size();
	
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetAdjacentObjects(const char * dst_type, UdmPseudoObjectS &value)
{
	CINT_TRY(dn_id, ob_id);
	
	Uml::Class kind = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), dst_type);
	if (!kind) throw udm_exception(string("No such kind: ") + dst_type);

	set<Udm::Object> ret = upi_o.GetAdjacentObjects(kind);
	set<Udm::Object>::iterator ret_i = ret.begin();
	int i = 0;
	while (ret_i != ret.end())
	{
		Udm::Object adj_ob = *ret_i;
		value.SetAt(i++, UdmPseudoObject(adj_ob.__impl()->__getdn()->uniqueId(), adj_ob.__impl()->uniqueId()));
	}
	
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetAdjacentObjectsCount(const char * dst_type, const AssociationInfo & ass, UdmLong &value) // HN5
{
	CINT_TRY(dn_id, ob_id);

	Uml::Class kind = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), dst_type);
	//kind can be null!

	Class o_ass_info_ass_class;

	if (ass.assoc_class)
	{
		o_ass_info_ass_class = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), ass.assoc_class.buffer());
		if (!o_ass_info_ass_class) throw udm_exception(string("No such association class : ") + ass.assoc_class.buffer());
	}
	Udm::Object::AssociationInfo o_ass_info(o_ass_info_ass_class);
	o_ass_info.strDstRoleName = ass.DstRolename.buffer();
	o_ass_info.strSrcRoleName = ass.SrcRolename.buffer();

	set<Udm::Object> ret = upi_o.GetAdjacentObjects(kind, o_ass_info);
	value.longVal = ret.size();
	
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetAdjacentObjects(const char * dst_type, const AssociationInfo & ass, UdmPseudoObjectS & value)
{
	CINT_TRY(dn_id, ob_id);

	Uml::Class kind = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), dst_type);
	//kind can be null!

	Class o_ass_info_ass_class;

	if (ass.assoc_class)
	{
		o_ass_info_ass_class = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), ass.assoc_class.buffer());
		if (!o_ass_info_ass_class) throw udm_exception(string("No such association class : ") + ass.assoc_class.buffer());
	}
	Udm::Object::AssociationInfo o_ass_info(o_ass_info_ass_class);
	o_ass_info.strDstRoleName = ass.DstRolename.buffer();
	o_ass_info.strSrcRoleName = ass.SrcRolename.buffer();

	set<Udm::Object> ret = upi_o.GetAdjacentObjects(kind, o_ass_info);
	set<Udm::Object>::iterator ret_i = ret.begin();
	int i = 0;
	while (ret_i != ret.end())
	{
		Udm::Object adj_ob = *ret_i;
		value.SetAt(i++, UdmPseudoObject(adj_ob.__impl()->__getdn()->uniqueId(), adj_ob.__impl()->uniqueId()));
	}
	
	
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetParent(UdmPseudoObject &value)
{
	return getParent(NULL, value);
};

bool UdmPseudoObject::GetChildObjectsCount(const char * type, UdmLong &value) // HN6
{
	return getChildrenCount(NULL, type, value);
};

bool UdmPseudoObject::GetChildObjects(const char * type, UdmPseudoObjectS &value)
{
	return getChildren(NULL, type, value);
};

bool UdmPseudoObject::GetChildObjectsCount(const CompositionInfo &ci, const cint_string & type, UdmLong &value) // HN7
{
	CINT_TRY(dn_id, ob_id);

	Udm::Object::CompositionInfo o_ci;
	o_ci.strChildRoleName = ci.childRole;
	o_ci.strParentRoleName = ci.parentRole;

	Class kind = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), type.buffer());

	set<Udm::Object> ret = upi_o.GetChildObjects(o_ci, kind);
	value.longVal = ret.size();
	
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetChildObjects(const CompositionInfo &ci, const cint_string & type, UdmPseudoObjectS &value)
{
	CINT_TRY(dn_id, ob_id);

	Udm::Object::CompositionInfo o_ci;
	o_ci.strChildRoleName = ci.childRole;
	o_ci.strParentRoleName = ci.parentRole;

	Class kind = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), type.buffer());

	set<Udm::Object> ret = upi_o.GetChildObjects(o_ci, kind);

	set<Udm::Object>::iterator ret_i = ret.begin();
	int i = 0;
	while (ret_i != ret.end())
	{
		Udm::Object child_ob = *ret_i;
		value.SetAt(i++, UdmPseudoObject(child_ob.__impl()->__getdn()->uniqueId(), child_ob.__impl()->uniqueId()));
	}
	
	return true;
	
	CINT_CATCH;
};

bool UdmPseudoObject::GetAssociationClassObjectsCount(const UdmPseudoObject &dst, const AssociationInfo & ass, UdmLong &value) // HN8
{
	CINT_TRY(dn_id, ob_id);

	Udm::Object dstObject = GetUdmObject(_upi_o_id(dst.dn_id, dst.ob_id));
	
	Class o_ass_info_ass_class;

	if (ass.assoc_class)
	{
		o_ass_info_ass_class = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), ass.assoc_class.buffer());
		if (!o_ass_info_ass_class) throw udm_exception(string("No such association class : ") + ass.assoc_class.buffer());
	}
	
	Udm::Object::AssociationInfo o_ass_info(o_ass_info_ass_class);
	o_ass_info.strDstRoleName = ass.DstRolename.buffer();
	o_ass_info.strSrcRoleName = ass.SrcRolename.buffer();

	set<Udm::Object> ret = upi_o.GetAssociationClassObjects(dstObject, o_ass_info);
	value.longVal = ret.size();
	
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetAssociationClassObjects(const UdmPseudoObject &dst, const AssociationInfo & ass, UdmPseudoObjectS &value)
{
	CINT_TRY(dn_id, ob_id);

	Udm::Object dstObject = GetUdmObject(_upi_o_id(dst.dn_id, dst.ob_id));
	
	Class o_ass_info_ass_class;

	if (ass.assoc_class)
	{
		o_ass_info_ass_class = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), ass.assoc_class.buffer());
		if (!o_ass_info_ass_class) throw udm_exception(string("No such association class : ") + ass.assoc_class.buffer());
	}
	
	Udm::Object::AssociationInfo o_ass_info(o_ass_info_ass_class);
	o_ass_info.strDstRoleName = ass.DstRolename.buffer();
	o_ass_info.strSrcRoleName = ass.SrcRolename.buffer();

	set<Udm::Object> ret = upi_o.GetAssociationClassObjects(dstObject, o_ass_info);
	set<Udm::Object>::iterator ret_i = ret.begin();
	int i = 0;
	while (ret_i != ret.end())
	{
		Udm::Object ass_class_ob = *ret_i;
		value.SetAt(i++, UdmPseudoObject(ass_class_ob.__impl()->__getdn()->uniqueId(), ass_class_ob.__impl()->uniqueId()));
	}
	
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetPeersFromAssociationClassObjectCount(UdmLong &value) // HN9
{
	CINT_TRY(dn_id, ob_id);

	// FIXME: Himanshu: Line below may not be required (left it there just for validation)
	pair<Udm::Object, Udm::Object> ret = upi_o.GetPeersFromAssociationClassObject();


	value.longVal = 2; // It is always a pair; above code is for call validation
	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::GetPeersFromAssociationClassObject(UdmPseudoObjectS &value)
{
	CINT_TRY(dn_id, ob_id);
	
	pair<Udm::Object, Udm::Object> ret = upi_o.GetPeersFromAssociationClassObject();
	int i = 0;
	
	Udm::Object peer_ob;

	peer_ob = ret.first;
	value.SetAt(i++, UdmPseudoObject(peer_ob.__impl()->__getdn()->uniqueId(), peer_ob.__impl()->uniqueId()));
	
	peer_ob = ret.second;
	value.SetAt(i++, UdmPseudoObject(peer_ob.__impl()->__getdn()->uniqueId(), peer_ob.__impl()->uniqueId()));
	
	
	return true;
	CINT_CATCH;
};

// Helper function
Uml::CompositionChildRole GetCompositionChildRoleByName(const Uml::Class& kind, const string strRoleName)
{
	set<Uml::CompositionChildRole> ccrs = CompositionPeerChildRoles(kind);
	set<Uml::CompositionChildRole>::iterator it = ccrs.begin();
	
	while(it != ccrs.end()) {
		Uml::CompositionChildRole aCCR = *it;
		if(string(aCCR.name()) == strRoleName) {
			return aCCR;
		}
		it++;
	}
	throw udm_exception(string("No such composition role found : ") + strRoleName);
}

// Creates a new child for 'this' object
bool UdmPseudoObject::CreateObject(const char * childType, UdmPseudoObject &newChild, const char* compositionChildRole, const bool subtype)
{
	CINT_TRY(dn_id, ob_id);

	Uml::Class childKind = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), childType);
	if (!childKind) throw udm_exception(string("No such kind : ") + childType);

	Udm::Object new_o;

	if(compositionChildRole != 0) {
		Uml::CompositionChildRole ccr = GetCompositionChildRoleByName(upi_o.type(), compositionChildRole);
		new_o = upi_o.Create(childKind, upi_o, ccr, 0, subtype);
	} else {
		new_o = upi_o.CreateObject(childKind);
	}

	newChild.dn_id = new_o.__impl()->__getdn()->uniqueId();
	newChild.ob_id = new_o.__impl()->uniqueId();

	return true;
	CINT_CATCH;
};

bool UdmPseudoObject::CreateLink(UdmPseudoObject& dst, const AssociationInfo& ass)
{
	CINT_TRY(dn_id, ob_id);

	Udm::Object dstObject = GetUdmObject(_upi_o_id(dst.dn_id, dst.ob_id));
	
	Class o_ass_info_ass_class;

	if (ass.assoc_class)
	{
		o_ass_info_ass_class = classByName(upi_o.__impl()->__getdn()->GetRootMeta(), ass.assoc_class.buffer());
		if (!o_ass_info_ass_class) throw udm_exception(string("No such association class : ") + ass.assoc_class.buffer());;
	}
	
	Udm::Object::AssociationInfo o_ass_info(o_ass_info_ass_class);
	o_ass_info.strDstRoleName = ass.DstRolename.buffer();
	o_ass_info.strSrcRoleName = ass.SrcRolename.buffer();

	return upi_o.CreateLink(dstObject, o_ass_info);
	
	CINT_CATCH;
};

bool UdmPseudoObject::DeleteObject()
{
	return detach();
};



UdmPseudoDataNetwork::UdmPseudoDataNetwork(const char * mn, UdmBool& meta_found)
{
	dn_id = -1;		//not opened yet
	try
	{
		//just to check if exists
		const Udm::UdmDiagram& dgr = Udm::MetaDepository::LocateDiagram(mn);
		dgr.init();
		meta_name = mn;
		meta_found.boolVal = true;

	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
		meta_found.boolVal = false;
	};
	//we have to make sure that diagram is a Uml::Diagram
}
//======================================================================
bool UdmPseudoDataNetwork::CreateNew(const char * sys_name, const char * meta_locator, const UdmPseudoObject& rootclass, const int backendsemantics)
{
	try
	{

		//at this point we don't have a data network ID
		if (dn_id != -1) throw udm_exception("UdmPseudoDataNetwork already opened!");
		//make sure root class is a Uml::Class and it is valid

		const Udm::UdmDiagram &dgr = Udm::MetaDepository::LocateDiagram(meta_name.buffer());

			
		Uml::Class rc;
		CINT_TRY(rootclass._dn_id(), rootclass._ob_id());
		rc = Uml::Class::Cast(upi_o);
		CINT_CATCH;
		
		
		//init dgr!
		UdmDom::DomDataNetwork * ddn = new UdmDom::DomDataNetwork(dgr);
		Udm::BackendSemantics ubs;
		switch (backendsemantics)
		{
		case UPDN_CHANGES_LOST_DEFAULT:
			ubs = CHANGES_LOST_DEFAULT;
			break;
		case UPDN_CHANGES_PERSIST_ALWAYS:
			ubs = CHANGES_PERSIST_ALWAYS;
			break;
		case UPDN_CHANGES_PERSIST_DEFAULT:
			ubs = CHANGES_PERSIST_DEFAULT;
			break;
		}
		ddn->CreateNew(sys_name, meta_locator, rc, ubs);
		dn_id = ddn->uniqueId();
		
		Udm::Object root_o = ddn->GetRootObject();
		if (!root_o) throw udm_exception("DataNetwork Creation failed!");

		UdmPseudoObject r(dn_id, root_o.__impl()->uniqueId());
		root = r;
		/*root.ob_id = root_o.__impl()->uniqueId();
		root.dn_id = dn_id;*/
		return true;	
	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
		return false;
	};

}
//======================================================================
bool UdmPseudoDataNetwork::CreateNewToString(const char * meta_locator, const UdmPseudoObject& rootclass)
{
	try
	{

		//at this point we don't have a data network ID
		if (dn_id != -1) throw udm_exception("UdmPseudoDataNetwork already opened!");
		//make sure root class is a Uml::Class and it is valid

		const Udm::UdmDiagram &dgr = Udm::MetaDepository::LocateDiagram(meta_name.buffer());

			
		Uml::Class rc;
		CINT_TRY(rootclass._dn_id(), rootclass._ob_id());
		rc = Uml::Class::Cast(upi_o);
		CINT_CATCH;
		
		
		//init dgr!
		UdmDom::DomDataNetwork * ddn = new UdmDom::DomDataNetwork(dgr);

		ddn->CreateNewToString(meta_locator, rc);
		dn_id = ddn->uniqueId();
		
		Udm::Object root_o = ddn->GetRootObject();
		if (!root_o) throw udm_exception("DataNetwork Creation failed!");

		UdmPseudoObject r(dn_id, root_o.__impl()->uniqueId());
		root = r;
		/*root.ob_id = root_o.__impl()->uniqueId();
		root.dn_id = dn_id;*/
		return true;	
	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
		return false;
	};

}

//======================================================================
bool UdmPseudoDataNetwork::OpenExisting(const char * sys_name, const char * meta_locator, const int backendsemantics)
{
	try
	{

		//at this point we don't have a data network ID
		if (dn_id != -1) throw udm_exception("UdmPseudoDataNetwork already opened!");
		//make sure root class is a Uml::Class and it is valid

		const Udm::UdmDiagram &dgr = Udm::MetaDepository::LocateDiagram(meta_name.buffer());

		//init dgr!
		UdmDom::DomDataNetwork * ddn = new UdmDom::DomDataNetwork(dgr);
		Udm::BackendSemantics ubs;
		switch (backendsemantics)
		{
		case UPDN_CHANGES_LOST_DEFAULT:
			ubs = CHANGES_LOST_DEFAULT;
			break;
		case UPDN_CHANGES_PERSIST_ALWAYS:
			ubs = CHANGES_PERSIST_ALWAYS;
			break;
		case UPDN_CHANGES_PERSIST_DEFAULT:
			ubs = CHANGES_PERSIST_DEFAULT;
			break;
		}
		ddn->OpenExisting(sys_name, meta_locator, ubs);
		dn_id = ddn->uniqueId();
		
		Udm::Object root_o = ddn->GetRootObject();
		if (!root_o) throw udm_exception("DataNetwork Creation failed!");

		/*root.ob_id = root_o.__impl()->uniqueId();
		root.dn_id = dn_id;
		*/

		UdmPseudoObject r(dn_id, root_o.__impl()->uniqueId());
		root = r;

		return true;	
	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
		return false;
	};

};
//===============================================================
bool UdmPseudoDataNetwork::OpenExistingFromString(const char * sys_name, const char * meta_locator)
{
	try
	{

		//at this point we don't have a data network ID
		if (dn_id != -1) throw udm_exception("UdmPseudoDataNetwork already opened!");
		//make sure root class is a Uml::Class and it is valid

		const Udm::UdmDiagram &dgr = Udm::MetaDepository::LocateDiagram(meta_name.buffer());

		//init dgr!
		UdmDom::DomDataNetwork * ddn = new UdmDom::DomDataNetwork(dgr);

    string  a(sys_name);
    string  b(meta_locator);
    ddn->OpenExistingFromString(a, b);
		dn_id = ddn->uniqueId();
		
		Udm::Object root_o = ddn->GetRootObject();
		if (!root_o) throw udm_exception("DataNetwork Creation failed!");

		/*root.ob_id = root_o.__impl()->uniqueId();
		root.dn_id = dn_id;
		*/

		UdmPseudoObject r(dn_id, root_o.__impl()->uniqueId());
		root = r;

		return true;	
	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
		return false;
	};

}
//================================================================
bool UdmPseudoDataNetwork::GetRootObject(UdmPseudoObject & value)
{
	value = root;
	return true;	
};
bool UdmPseudoDataNetwork::GetRootMeta(UdmPseudoObject &value)			//a Uml::Diagram should be returned
{
	CINT_DN_TRY(dn_id);
	const Uml::Diagram& dgr = dn->GetRootMeta();
	
	/*value.dn_id = dgr.__impl()->__getdn()->uniqueId();
	value.ob_id = dgr.__impl()->uniqueId();
	*/

	UdmPseudoObject r(dgr.__impl()->__getdn()->uniqueId(), dgr.__impl()->uniqueId());
	value = r;

	return true;	
	CINT_DN_CATCH
};
bool UdmPseudoDataNetwork::CloseWithUpdate()
{
	CINT_DN_TRY(dn_id);
	dn->CloseWithUpdate();
	return true;	
	CINT_DN_CATCH
};
bool UdmPseudoDataNetwork::CloseNoUpdate()
{
	CINT_DN_TRY(dn_id);
	dn->CloseNoUpdate();
	return true;	
	CINT_DN_CATCH
};
bool UdmPseudoDataNetwork::SaveAsString(cint_string& buffer)
{
	CINT_DN_TRY(dn_id);

  cint_string tmp(dn->Str().c_str());
  buffer = tmp;
	return true;	
	CINT_DN_CATCH
};

bool UdmPseudoDataNetwork::SaveAs(const char *systemname)
{
	CINT_DN_TRY(dn_id);
	dn->SaveAs(systemname);
	return true;	
	CINT_DN_CATCH
};
bool UdmPseudoDataNetwork::CloseAs(const char * systemname)
{
	CINT_DN_TRY(dn_id);
	dn->CloseAs(systemname);
	return true;	
	CINT_DN_CATCH
};
bool UdmPseudoDataNetwork::isOpen(UdmBool &value)
{
	CINT_DN_TRY(dn_id);
	value.boolVal = dn->isOpen();
	return true;	
	CINT_DN_CATCH
};
bool UdmPseudoDataNetwork::uniqueId(UdmLong &value)
{
	CINT_DN_TRY(dn_id);
	value.longVal = dn->uniqueId();
	return true;	
	CINT_DN_CATCH
};
UdmPseudoDataNetwork::~UdmPseudoDataNetwork()
{
	DataNetwork* dn = Udm::DataNetwork::GetDnById(dn_id);
	if (dn) delete dn;
	
};


bool UPO_SetDiagram(UdmPseudoObject &diagram_what, const char* name)
{
	try
	{
		const Udm::UdmDiagram & dgr = Udm::MetaDepository::LocateDiagram(name);
		/*diagram_what.dn_id = dgr.dgr->__impl()->__getdn()->uniqueId();
		diagram_what.ob_id = dgr.dgr->uniqueId();*/

		UdmPseudoObject r(dgr.dgr->__impl()->__getdn()->uniqueId(), dgr.dgr->uniqueId());
		diagram_what = r;
		return true;
	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
		return false;
	};
};
bool UPO_SetClass(UdmPseudoObject &class_what, UdmPseudoObject &diagram_what_dgr, const char* target_name)
{
//	CINT_TRY(dn_id, ob_id);

	CINT_TRY(diagram_what_dgr._dn_id(), diagram_what_dgr._ob_id());
	Uml::Diagram dgr = Uml::Diagram::Cast(upi_o);
	if (!dgr) throw udm_exception("The provided diagram is NULL!");

	Uml::Class what;
	SetClass(what, dgr, target_name);
	if (!what) throw udm_exception("SetClass() failed!");
/*	class_what.dn_id = what.__impl()->__getdn()->uniqueId();
	class_what.ob_id = what.__impl()->uniqueId();
	*/
	UdmPseudoObject r(what.__impl()->__getdn()->uniqueId(), what.__impl()->uniqueId());
	class_what = r;
	return true;
	CINT_CATCH;

};

bool UPO_SetAttribute(UdmPseudoObject &attr_what, UdmPseudoObject &class_what_class,  const char* target_name)
{
	CINT_TRY(class_what_class._dn_id(), class_what_class._ob_id());
	Uml::Class what_class = Uml::Class::Cast(upi_o);
	if (!what_class) throw udm_exception("The provided class is NULL!");

	Uml::Attribute what;
	SetAttribute(what, what_class, target_name);
	if (!what) throw udm_exception("SetAttribute failed!");
/*	attr_what.dn_id = what.__impl()->__getdn()->uniqueId();
	attr_what.ob_id = what.__impl()->uniqueId();
	*/
	UdmPseudoObject r(what.__impl()->__getdn()->uniqueId(), what.__impl()->uniqueId());
	attr_what = r;
	return true;
	CINT_CATCH;
};

bool UPO_SetChildRole(UdmPseudoObject &ccr_what, UdmPseudoObject &what_target_class, UdmPseudoObject &what_theo_target_class, const char * role_name, const char * orole_name)
{
	CINT_TRY2(what_target_class._dn_id(), what_target_class._ob_id(), what_theo_target_class._dn_id(), what_theo_target_class._ob_id());
	Uml::Class what_target = Uml::Class::Cast(upi_o1);
	if (!what_target) throw udm_exception("The provided class is NULL!");

	Uml::Class what_theo_target = Uml::Class::Cast(upi_o2);
	if (!what_theo_target) throw udm_exception("The provided class is NULL!");

	Uml::CompositionChildRole what;
	SetChildRole(what,what_target, what_theo_target, role_name, orole_name);

	if (!what) throw udm_exception("SetChildRole failed!");
	/*
	ccr_what.dn_id = what.__impl()->__getdn()->uniqueId();
	ccr_what.ob_id = what.__impl()->uniqueId();
	*/

	UdmPseudoObject r(what.__impl()->__getdn()->uniqueId(), what.__impl()->uniqueId());
	ccr_what = r;
	
	return true;
	CINT_CATCH;
};
bool UPO_SetParentRole(UdmPseudoObject &cpr_what, UdmPseudoObject &what_target_class, UdmPseudoObject &what_theo_target_class, const char * role_name, const char* orole_name)
{
	CINT_TRY2(what_target_class._dn_id(), what_target_class._ob_id(), what_theo_target_class._dn_id(), what_theo_target_class._ob_id());
	Uml::Class what_target = Uml::Class::Cast(upi_o1);
	if (!what_target) throw udm_exception("The provided class is NULL!");

	Uml::Class what_theo_target = Uml::Class::Cast(upi_o2);
	if (!what_theo_target) throw udm_exception("The provided class is NULL!");

	Uml::CompositionParentRole what;
	SetParentRole(what,what_target, what_theo_target, role_name, orole_name);

	if (!what) throw udm_exception("SetChildRole failed!");
	/*
	cpr_what.dn_id = what.__impl()->__getdn()->uniqueId();
	cpr_what.ob_id = what.__impl()->uniqueId();
	*/
	UdmPseudoObject r(what.__impl()->__getdn()->uniqueId(), what.__impl()->uniqueId());
	cpr_what = r;
	
	return true;
	CINT_CATCH;
};
bool UPO_SetAssocRole(UdmPseudoObject &ar_what, UdmPseudoObject &what_target_class, UdmPseudoObject &what_theo_target_class, const char* target_name)
{
	CINT_TRY2(what_target_class._dn_id(), what_target_class._ob_id(), what_theo_target_class._dn_id(), what_theo_target_class._ob_id());
	Uml::Class what_target = Uml::Class::Cast(upi_o1);
	if (!what_target) throw udm_exception("The provided class is NULL!");

	Uml::Class what_theo_target = Uml::Class::Cast(upi_o2);
	if (!what_theo_target) throw udm_exception("The provided class is NULL!");

	Uml::AssociationRole what;
	SetAssocRole(what,what_target, what_theo_target, target_name);

	if (!what) throw udm_exception("SetChildRole failed!");
	/*
	ar_what.dn_id = what.__impl()->__getdn()->uniqueId();
	ar_what.ob_id = what.__impl()->uniqueId();
	*/
	UdmPseudoObject r(what.__impl()->__getdn()->uniqueId(), what.__impl()->uniqueId());
	ar_what = r;
	return true;
	CINT_CATCH;
};
void UPO_dummy(){};

class UPO_metainfo
{
	Udm::DataNetwork * dn;
	Uml::Diagram * udgr;
	Udm::UdmDiagram * dgr;
public:
	UPO_metainfo(Udm::DataNetwork *_dn, Uml::Diagram * _udgr, Udm::UdmDiagram * _dgr) : dn(_dn), udgr(_udgr), dgr(_dgr) {} ;
	~UPO_metainfo()
	{
		Udm::MetaDepository::RemoveDiagram(udgr->name());

		delete udgr, udgr=0;
		delete dgr,dgr=0;
		delete dn,dn=0;
	};
static map<string,const UPO_metainfo *> upo_metamap;

static bool reg_metainfo(const string & str, const UPO_metainfo * mi)
	{
		map<string, const UPO_metainfo *>::value_type item(str, mi); 
		pair<map<string, const UPO_metainfo *>::iterator, bool> res = upo_metamap.insert(item);
		return res.second;
	};
static const UPO_metainfo * get_metainfo(const string &str)
	{
		map<string, const UPO_metainfo *>::iterator i = upo_metamap.find(str);
		if (i != upo_metamap.end()) return i->second;
		else return NULL;
	};

static void unreg_metainfo(const string &str)
	{
		upo_metamap.erase(str);
	};
	
};

map<string,const UPO_metainfo *> UPO_metainfo::upo_metamap;



bool UDM_DLL UPO_LoadDiagram(const char * xml_meta_file, UdmPseudoObject & diagram)
{
	try
	{
		//we get an UML XML file 
		//we have to load it and register it with MetaDepository
		//somewhere we have to release these objects
		UdmDom::DomDataNetwork * ddn = new UdmDom::DomDataNetwork(Uml::diagram);
		Uml::Diagram * loaded_dgr = new Uml::Diagram();
		Udm::UdmDiagram * dgr = new Udm::UdmDiagram();	//this will be fried by RemoveDiagram();

		ddn->OpenExisting(xml_meta_file, "Uml.xsd", Udm::CHANGES_LOST_DEFAULT);
		*loaded_dgr = Uml::Diagram::Cast(ddn->GetRootObject());

		if (!loaded_dgr) throw udm_exception("UML Diagram could not be loaded!");
		
		const UPO_metainfo * umi = new UPO_metainfo(ddn, loaded_dgr, dgr);
		if (!UPO_metainfo::reg_metainfo(xml_meta_file, umi))
			throw udm_exception("MetaInfo is already registered, please unload first.");


		dgr->init = UPO_dummy;
		dgr->dgr = loaded_dgr;
		
		Udm::MetaDepository::StoreDiagram(loaded_dgr->name(),*dgr);

		//return value set-up
		/*
		diagram.dn_id = ddn->uniqueId();
		diagram.ob_id = loaded_dgr->__impl()->uniqueId();
		*/
		UdmPseudoObject r(ddn->uniqueId(), loaded_dgr->__impl()->uniqueId());
		diagram = r;
		return true;
	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
		return false;
	};
}

bool UDM_DLL UPO_LoadDiagramFromString(const char * xml_meta_file, const char * xml_string, UdmPseudoObject & diagram)
{
	try
	{
		//we get an UML XML file 
		//we have to load it and register it with MetaDepository
		//somewhere we have to release these objects
		UdmDom::DomDataNetwork * ddn = new UdmDom::DomDataNetwork(::Uml::diagram);
		Uml::Diagram * loaded_dgr = new Uml::Diagram();
		Udm::UdmDiagram * dgr = new Udm::UdmDiagram();	//this will be fried by RemoveDiagram();

    string xml(xml_string);
		ddn->OpenExistingFromString(xml, "Uml.xsd");
		*loaded_dgr = Uml::Diagram::Cast(ddn->GetRootObject());

		if (!loaded_dgr) throw udm_exception("UML Diagram could not be loaded!");
		
		const UPO_metainfo * umi = new UPO_metainfo(ddn, loaded_dgr, dgr);
		if (!UPO_metainfo::reg_metainfo(xml_meta_file, umi))
			throw udm_exception("MetaInfo is already registered, please unload first.");


		dgr->init = UPO_dummy;
		dgr->dgr = loaded_dgr;
		
		Udm::MetaDepository::StoreDiagram(loaded_dgr->name(),*dgr);

		//return value set-up
		/*
		diagram.dn_id = ddn->uniqueId();
		diagram.ob_id = loaded_dgr->__impl()->uniqueId();
		*/
		UdmPseudoObject r(ddn->uniqueId(), loaded_dgr->__impl()->uniqueId());
		diagram = r;
		return true;
	}
	catch (udm_exception &e)
	{
		set_last_error(e.what());
		return false;
	};
};


bool UDM_DLL UPO_UnLoadDiagram(const char * xml_meta_file)
{
	const UPO_metainfo * umi = UPO_metainfo::get_metainfo(xml_meta_file);

	if (!umi) return false;
		
	UPO_metainfo::unreg_metainfo(xml_meta_file);
	delete umi;
	
	return true;

};

//===========================
bool UDM_DLL StoreXsd(const cint_string& key, const cint_string& xsd_str)
{
  try
  {
    UdmDom::str_xsd_storage::StoreXsd((const std::string&)key.buffer(), (const std::string&)xsd_str.buffer());
  }
  catch (const udm_exception& e) 
  {
    set_last_error(e.what()); return false;
  } 
  catch(...) 
  {
    set_last_error("Unexpected exception caught."); 
    return false;
  }
  return true;
}
//===========================
		
bool UDM_DLL RemoveXsd(const cint_string& key)
{
  try
  {

    UdmDom::str_xsd_storage::RemoveXsd((const std::string)key.buffer());
  }
  catch (const udm_exception& e) 
  {
    set_last_error(e.what()); return false;
  } 
  catch(...) 
  {
    set_last_error("Unexpected exception caught."); 
    return false;
  }

  return true;
}
bool UDM_DLL ClearXsdStorage()
{
  try
  {

    UdmDom::str_xsd_storage::ClearXsdStorage();
  }
  catch (const udm_exception& e) 
  {
    set_last_error(e.what()); return false;
  } 
  catch(...) 
  {
    set_last_error("Unexpected exception caught."); 
    return false;
  }

  return true;
}
